@@ -9,7 +9,9 @@ def main():
     import sys
     import time
     import signal
+    import tempfile
 
+    import yara_scanner
     from yara_scanner import YaraScannerServer
 
     parser = argparse.ArgumentParser(description="Yara Scanner Server")
@@ -48,6 +50,34 @@ def main():
         type=int,
         help="How often to check for modifications to the yara rules (in seconds). Defaults to 60.",
     )
+    parser.add_argument(
+        "-m",
+        "--max-workers",
+        required=False,
+        default=None,
+        type=int,
+        help="Specify the maximum number of workers. Defaults to the local cpu count.",
+    )
+    parser.add_argument(
+        "--cache-directory",
+        required=False,
+        default=None,
+        help="The directory to use to cache compiled yara programs. Defaults to the system temp dir.",
+    )
+    parser.add_argument(
+        "--max-contexts",
+        required=False,
+        default=yara_scanner.DEFAULT_MAX_CONTEXTS,
+        type=int,
+        help=f"How many yara contexts can stay in memory at one time. Applies to prefiltering. Defaults to {yara_scanner.DEFAULT_MAX_CONTEXTS}.",
+    )
+    parser.add_argument(
+        "--disable-prefilter",
+        required=False,
+        action="store_true",
+        default=False,
+        help="Disable pre-filtering.",
+    )
     parser.add_argument(
         "--pid-file",
         required=False,
@@ -72,6 +102,21 @@ def main():
         action="store_true",
         help="Kill the currently executing yara scanner server.",
     )
+    # resource constraints
+    parser.add_argument(
+        "-T",
+        "--timeout",
+        default=yara_scanner.DEFAULT_TIMEOUT,
+        type=int,
+        help="""Maximum amount of time (in seconds) a single scan is allowed to take.
+        Passed directly to libyara.""")
+
+    parser.add_argument(
+        "-M",
+        "--max-bytes",
+        default=yara_scanner.DEFAULT_MAX_BYTES,
+        type=int,
+        help="""Only the first N bytes of a file are scanned.""")
     args = parser.parse_args()
 
     if not os.path.isdir(args.base_dir):
@@ -208,12 +253,22 @@ def main():
         os.dup2(0, 1)
         os.dup2(0, 2)
 
+    context_cache_dir = args.cache_directory
+    if context_cache_dir is None:
+        context_cache_dir = os.path.join(tempfile.gettempdir(), ".yss_cache")
+
     server = YaraScannerServer(
         base_dir=args.base_dir,
         signature_dir=args.signature_dir,
         socket_dir=args.socket_dir,
         update_frequency=args.update_frequency,
         backlog=args.backlog,
+        default_timeout=args.timeout,
+        max_bytes=args.max_bytes,
+        max_workers=args.max_workers,
+        disable_prefilter=args.disable_prefilter,
+        context_cache_dir=context_cache_dir,
+        max_contexts=args.max_contexts,
     )
 
     try:
